<form class="form-horizontal form-complex">
{% if mod_alert.get("parameters") %}
	{% for param in mod_alert.parameters %}
		{% with mod_alert=mod_alert, param=param %}
			{% include param.format_type + ".html" %}
		{% endwith %}
	{% endfor %}
{% endif %}
</form>
