<!--
  ~ Copyright 2023 Splunk Inc.
  ~
  ~ Licensed under the Apache License, Version 2.0 (the "License");
  ~ you may not use this file except in compliance with the License.
  ~ You may obtain a copy of the License at
  ~
  ~ http://www.apache.org/licenses/LICENSE-2.0
  ~
  ~ Unless required by applicable law or agreed to in writing, software
  ~ distributed under the License is distributed on an "AS IS" BASIS,
  ~ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  ~ See the License for the specific language governing permissions and
  ~ limitations under the License.
  ~
-->
<%! app_name = cherrypy.request.path_info.split('/')[3] %>\
<!DOCTYPE html>
<html class="no-js" lang="">
    <head>
        <meta charset="utf-8" />
        <meta http-equiv="x-ua-compatible" content="ie=edge" />
        <title>${_('Loading')}</title>
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <link rel="apple-touch-icon" href="apple-touch-icon.png" />
    </head>

    <body>
        <script src="${make_url('/config?autoload=1')}" crossorigin="use-credentials"></script>
        <script src="${make_url('/static/js/i18n.js')}"></script>
        <script src="${make_url('/i18ncatalog?autoload=1')}"></script>
        <script>
            __splunkd_partials__ = ${json_decode(splunkd)};
        </script>

        <% page_path = "/static/app/" + app_name + "/js/build/entry_page.js" %>

        <script src="${make_url(page_path)}"></script>
    </body>
</html>
